# Created June 2002 by Mohammad Hassan.
# routing generator class
Class PolicyGen

proc policy-usage { } {
puts {usage: routing [-<key 1> <value 1> ... -<key n> <value n>]

example options: 
-outfile f -policer TokenBucket -q1min 20 -q1max 40 -q1prob 0.2

keys and corresponding values:
-policer possible values: TokenBucket, TSW2CM, TSW3CM, srTCM, trTCM
                          All policcers have a initial code point of 10 
                          and downgraded codepoints of 11 and 12.
                          code point for unmarked traffic is 0.
                          At all links 
                                   code point 10 is mapped to queue (0,0) i.e queue1
                                   code point 11 is mapped to queue (0,1) i.e queue2
                                   code point 12 is mapped to queue (0,2) i.e queue3
                                   code point 0 is mapped to queue (1,0)  i.e queue4
    
-q1min, -q2min, -q3min, -q4min [min values for queue number 1,2,3,4 respectively]
-q1max, -q2max, -q3max, -q4max [max values for queue number 1,2,3,4 respectively]
-q1prob, -q2prob, -q3prob, -q4prob [drop probablity for queue number 1,2,3,4 respectively]

-qosr possible values: 1,0
                        set to "1" if you want to use qosr routing with diffserv
                        ELSE set to 0 
                                  
q1 (queue1) is queue (0,0)
q2 (queue2) is queue (0,1)
q3 (queue3) is queue (0,2)
q4 (queue4) is queue (1,0)

CIR PIR CBS EBS and PBS are not taken as input, but default values are used for them.
to change their values go to the policy tcl file generated by this generator.

VERY IMP: go to the topo file generated by the topo generator 
          and (since the diffserv in ns-2 does not work for heirarichal topology)
          1) replace the line "set n($i) [$ns node $a]" by "set n($i) [$ns node]"
          2) delete the folowing lines in it
                            lappend domain 1
                            AddrParams set domain_num_ $domain
                            lappend cluster 34
                            AddrParams set cluster_num_ $cluster
                            lappend eilastlevel 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
                            AddrParams set nodes_num_ $eilastlevel

PLEASE GO THROUGH THE HELP OF mtransit_stub TOPOLOGY GENERATED BY THE "topology -h 4" COMMAND
TO KNOW ABOUT THE PRECAUTONS TO BE TAKEN BEFORE GENERATING A TOPLOGY TO BE USED BY policy-gen.tcl
     
  }
}



proc policy { args } {
	set len [llength $args]

	if { $len } {
	    set key [lindex $args 0]
            if {$key == "-?" || $key == "--help" || $key == "-help" \
			|| $key == "-h" } {
			policy-usage
			return
                }
	}

        if { [expr $len % 2] } {
                # if number is odd => error !
                puts "fewer number of arguments than needed in \"$args\""
                policy-usage
		return
        }

	# check if the PolicyGen type exists
	if { [catch {set rg [PolicyGen info instances]}] } {
		puts "no class PolicyGen"
		policy-usage
		return
	}
	if { $rg == "" } {
		set rg [new PolicyGen]
	}
	if { ![llength $args] } {
		$rg create
	} else {
		$rg create $args
	}
}

PolicyGen instproc init { } {
	$self next
}

PolicyGen instproc default_options { } {

	$self instvar opt_info

	set opt_info {
		# init file to -1, must be supplied by input
		outfile -1

		# setting default red parametres
		q1min 20
                q1max 40
                q1prob 0.02

                q2min 10
                q2max 20
                q2prob 0.10

                q3min 5
                q3max 10
                q3prob 0.20

                q4min 3
                q4max 6
                q4prob 0.30

		# default policer type
		policer TokenBucket

		# defalult CIR CBS PIR....ect parametres
                # should be changed manually before runnig the simulation

                CIR 2000000
                PIR 3000000
                CBS 1000
                EBS 2000
                PBS 1000

                qosr 0


	}
	$self parse_opts
}
	
# are the parsing functions general enuf... ! chk
PolicyGen instproc parse_opts { } {
	$self instvar opts opt_info

	while { $opt_info != ""} {
		# parse line by line
                if {![regexp "^\[^\n\]*\n" $opt_info line]} {
                        break  
                }
		# remove the parsed line
                regsub "^\[^\n\]*\n" $opt_info {} opt_info
		# remove leading spaces and tabs using trim
                set line [string trim $line]
		# skip comment lines beginning with #
                if {[regexp "^\[ \t\]*#" $line]} {
                        continue
                }
		# skip empty lines
                if {$line == ""} {
                        continue
                } elseif [regexp {^([^ ]+)[ ]+([^ ]+)$} $line dummy key value] {
                        set opts($key) $value
                } 
	}
}

PolicyGen instproc parse_input { args } {
	# remove the list brackets from the args list
        set args [lindex $args 0]
        set len [llength $args]

	$self instvar opts

	for { set i 0 } { $i < $len } { incr i } {
		set key [lindex $args $i]
		regsub {^-} $key {} key
                if {![info exists opts($key)]} {
			puts stderr "unrecognized option $key"
			policy-usage
			return -1
		}
		incr i
		# puts "changing $key from $opts($key) to [lindex $args $i]"
		set opts($key) [lindex $args $i]
	}
	# puts "end of parsing... "
	return 0
}

PolicyGen instproc create { args } {
        # remove the list brackets from the args list
        set args [lindex $args 0]
        set len [llength $args]
        # puts "calling create with args $args, len $len"

	$self default_options

	if { $len } {
		if { [$self parse_input $args] == -1 } {
			return 
		}
	}
	# check that the filename is provided
	$self instvar opts
	if { $opts(outfile) == -1 } {
	 puts {you must provide outfile name. use "routing -h" for help}
		return
	}
	$self create-policy
}

PolicyGen instproc create-policy { } {
	# XXX leave out the checks to ns later.. !
	# this would allow extensibility to include other kinds of 
	# routing without having to modify the generator
	# if { [$self input-check] == -1 } {
	#	puts "There's an input error !! aborting"
	#	policy-usage
	# 	return
	# }
	
	$self instvar opts

	set file $opts(outfile)

	$self generate-code $file
}


PolicyGen instproc GetPol { type } {  

         if {[catch {set tg [ScenGen getTG]}]} {
            puts stderr {Can't find topology generator..quiting!
                        Either run the topology generator first, or provide the total
                        number of nodes using the "totalnodes" option.
                        Use "agents -h" for more help.}
            exit
           }
         return [$tg getPol $type]

}




PolicyGen instproc GetNodes { type } {

         if {[catch {set tg [ScenGen getTG]}]} {
            puts stderr {Can't find topology generator..quiting!
                        Either run the topology generator first, or provide the total
                        number of nodes using the "totalnodes" option.
                        Use "agents -h" for more help.}
            exit
           }
         return [$tg getNodes $type]        

}

PolicyGen instproc setdomainpolicy { f } {
$self instvar opts

       puts $f "\t\t set CIR0 $opts(CIR)"
       puts $f "\t\t set PIR0 $opts(PIR)"
       puts $f "\t\t set CBS0 $opts(CBS)"
       puts $f "\t\t set EBS0 $opts(EBS)"
       puts $f "\t\t set PBS0 $opts(PBS)"

set nodlist [$self GetNodes "allofcores"]
puts $f "\t foreach s1 \[split $nodlist :\] \{"

puts $f "\t foreach abc \{"
set edgelist [$self GetNodes "diffedges"]
foreach diedge [split $edgelist :] {
set corelist [$self GetNodes edgetocore$diedge]
          foreach dicore [split $corelist :] {
          puts $f "\t\t \{$diedge $dicore\}"  
          }

}
       puts $f "\} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"

       puts $f "\t\t switch $opts(policer) \{"
       puts $f "\t\t \"TokenBucket\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 -1 TokenBucket 10 \$CIR0 \$CBS0"
       puts $f "\t\t\t \$qEC addPolicerEntry TokenBucket 10 11"
       puts $f "\t\t \}"
       puts $f "\t\t \"TSW2CM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 -1 TSW2CM 10 \$CIR0"
       puts $f "\t\t\t \$qEC addPolicerEntry TSW2CM 10 11"
       puts $f "\t\t \}"
       puts $f "\t\t \"TSW3CM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 -1 TSW3CM 10 \$CIR0 \$PIR0"
       puts $f "\t\t\t \$qEC addPolicerEntry TSW3CM 10 11 12"
       puts $f "\t\t \}"
       puts $f "\t\t \"srTCM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 -1 srTCM 10 \$CIR0 \$CBS0 \$EBS0"
       puts $f "\t\t\t \$qEC addPolicerEntry srTCM 10 11 12"
       puts $f "\t\t \}"
       puts $f "\t\t \"trTCM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 -1 trTCM 10 \$CIR0 \$CBS0 \$PIR0 \$PBS"
       puts $f "\t\t\t \$qEC addPolicerEntry trTCM 10 11 12"
       puts $f "\t\t \}"

       puts $f "\t\t \}"
       puts $f "\t\t \}"
puts $f "\t\t \}"


}

PolicyGen instproc generate-code { file } {
        $self instvar opts
	set f [open $file w]

	# puts "generate-code 
        

	puts $f "proc setup-policy \{ sim node \} \{ "
	puts $f "\t upvar \$sim ns"
        puts $f "\t upvar \$node no"
        puts $f "\t \# Here you can change the values of CIR CBR ect....."
        puts $f "\t \# The Format is "
        puts $f "\t \# \{edge_node core_node source_node desination_node CIR PIR CBS EBS PBS\}"
        # Now set up policies at all the required edge to core links
        # Sets policies ony at those edge to core links through which traffic is passing

        puts $f "\t foreach abc \{"
              
 
        set edgelist [$self GetNodes "diffedges"]

        foreach diedge [split $edgelist :] {
          set trlist [$self GetPol policy$diedge]
          if {$trlist==-1} {
          set boo -1
          } else {
          set boo 1
          } 
          if {$boo==1} {   
          set corelist [$self GetNodes edgetocore$diedge]
          foreach tr [split $trlist :] {
              set sd [split $tr ?]
              set trs [lindex $sd 0]
              set trd [lindex $sd 1]
              foreach dicore [split $corelist :] {
              puts $f "\t\t \{$diedge $dicore $trs $trd $opts(CIR) $opts(PIR) $opts(CBS) $opts(EBS) $opts(PBS)\}"  
              }

              set destgetinslist [$self GetNodes getins$trd]
              foreach desgetins [split $destgetinslist :] {
               set destcoreslist [$self GetNodes edgetocore$desgetins]
                 foreach destcore [split $destcoreslist :] {
                 puts $f "\t\t \{$desgetins $destcore $trd $trs $opts(CIR) $opts(PIR) $opts(CBS) $opts(EBS) $opts(PBS)\}"
               }
              }
        }
        }
        }
        
       
       puts $f "\} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"
       puts $f "\t\t \$qEC set numQueues_ 2"
       puts $f "\t\t \$qEC setNumPrec 3"
       
       
       puts $f "\t\t set s1 \$no\(\[lindex \$abc 2\]\)"
       puts $f "\t\t set dest \$no\(\[lindex \$abc 3\]\)"
       puts $f "\t\t set CIR0 \[lindex \$abc 4\]"
       puts $f "\t\t set PIR0 \[lindex \$abc 5\]"
       puts $f "\t\t set CBS0 \[lindex \$abc 6\]"
       puts $f "\t\t set EBS0 \[lindex \$abc 7\]"
       puts $f "\t\t set PBS0 \[lindex \$abc 8\]"
       
       puts $f "\t\t set s1 \[\$s1 id\]"
       puts $f "\t\t set dest \[\$dest id\]"
       
       puts $f "\t\t switch $opts(policer) \{"
      
       puts $f "\t\t \"TokenBucket\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 \$dest TokenBucket 10 \$CIR0 \$CBS0"
       puts $f "\t\t\t \$qEC addPolicerEntry TokenBucket 10 11"       
   #    puts $f "\t\t\t \$qEC addPHBEntry 10 0 0"
   #    puts $f "\t\t\t \$qEC addPHBEntry 11 0 1"
   #    puts $f "\t\t\t \$qEC addPHBEntry 0 1 0"
       # puts $f "\t\t\t \$qEC printPolicyTable"
       puts $f "\t\t \}"
       
       puts $f "\t\t \"TSW2CM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 \$dest TSW2CM 10 \$CIR0"
       puts $f "\t\t\t \$qEC addPolicerEntry TSW2CM 10 11"
     #  puts $f "\t\t\t \$qEC addPHBEntry 10 0 0"
     #  puts $f "\t\t\t \$qEC addPHBEntry 11 0 1"
     #  puts $f "\t\t\t \$qEC addPHBEntry 0 1 0"
       puts $f "\t\t \}"
       
       puts $f "\t\t \"TSW3CM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 \$dest TSW3CM 10 \$CIR0 \$PIR0"
       puts $f "\t\t\t \$qEC addPolicerEntry TSW3CM 10 11 12"
     #  puts $f "\t\t\t \$qEC addPHBEntry 10 0 0"
     #  puts $f "\t\t\t \$qEC addPHBEntry 11 0 1"
     #  puts $f "\t\t\t \$qEC addPHBEntry 12 0 2"
     #  puts $f "\t\t\t \$qEC addPHBEntry 0 1 0"
       puts $f "\t\t \}"

       puts $f "\t\t \"srTCM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 \$dest srTCM 10 \$CIR0 \$CBS0 \$EBS0"
       puts $f "\t\t\t \$qEC addPolicerEntry srTCM 10 11 12"
     #  puts $f "\t\t\t \$qEC addPHBEntry 10 0 0"
     #  puts $f "\t\t\t \$qEC addPHBEntry 11 0 1"
     #  puts $f "\t\t\t \$qEC addPHBEntry 12 0 2"
     #  puts $f "\t\t\t \$qEC addPHBEntry 0 1 0"
       puts $f "\t\t \}" 
        
       puts $f "\t\t \"trTCM\" \{"
       puts $f "\t\t\t \$qEC addPolicyEntry \$s1 \$dest trTCM 10 \$CIR0 \$CBS0 \$PIR0 \$PBS"
       puts $f "\t\t\t \$qEC addPolicerEntry trTCM 10 11 12"
     #  puts $f "\t\t\t \$qEC addPHBEntry 10 0 0"
     #  puts $f "\t\t\t \$qEC addPHBEntry 11 0 1"
     #  puts $f "\t\t\t \$qEC addPHBEntry 12 0 2"
     #  puts $f "\t\t\t \$qEC addPHBEntry 0 1 0"
       puts $f "\t\t \}"
       
       puts $f "\t\t \}"
 

     #  puts $f "\t\t \$qEC configQ 0 0 $opts(q1min) $opts(q1max) $opts(q1prob)"
     #  puts $f "\t\t \$qEC configQ 0 1 $opts(q2min) $opts(q2max) $opts(q2prob)"
     #  puts $f "\t\t \$qEC configQ 0 2 $opts(q3min) $opts(q3max) $opts(q3prob)"
     #  puts $f "\t\t \$qEC configQ 1 0 $opts(q4min) $opts(q4max) $opts(q4prob)"

       puts $f "\t \}"

       if {$opts(qosr)==1} {
        $self setdomainpolicy $f
       }

      # set PHB and queue parametres at all the links 
       
       puts $f "\t foreach abc \{"
          
        
        set edgelist [$self GetNodes "diffedges"]

        foreach diedge [split $edgelist :] {
          set corelist [$self GetNodes "edgetocore$diedge"]
          foreach dicore [split $corelist :] {
          puts $f "\t\t \{$dicore $diedge\}"
          }
        }
 
       puts $f "\t \} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"
       puts $f "\t\t \$qEC set numQueues_ 2"
       puts $f "\t\t \$qEC setNumPrec 3"


       puts $f "\t\t \$qEC configQ 0 0 $opts(q1min) $opts(q1max) $opts(q1prob)"
       puts $f "\t\t \$qEC configQ 0 1 $opts(q2min) $opts(q2max) $opts(q2prob)"
       puts $f "\t\t \$qEC configQ 0 2 $opts(q3min) $opts(q3max) $opts(q3prob)"
       puts $f "\t\t \$qEC configQ 1 0 $opts(q4min) $opts(q4max) $opts(q4prob)"

       
       puts $f "\t\t \$qEC addPHBEntry 10 0 0"
       puts $f "\t\t \$qEC addPHBEntry 11 0 1"
       puts $f "\t\t \$qEC addPHBEntry 12 0 2"
       puts $f "\t\t \$qEC addPHBEntry 0 1 0"


       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 1\]\) \$no\(\[lindex \$abc 0\]\)\] queue\]"
       puts $f "\t\t \$qEC set numQueues_ 2"
       puts $f "\t\t \$qEC setNumPrec 3"


       puts $f "\t\t \$qEC configQ 0 0 $opts(q1min) $opts(q1max) $opts(q1prob)"
       puts $f "\t\t \$qEC configQ 0 1 $opts(q2min) $opts(q2max) $opts(q2prob)"
       puts $f "\t\t \$qEC configQ 0 2 $opts(q3min) $opts(q3max) $opts(q3prob)"
       puts $f "\t\t \$qEC configQ 1 0 $opts(q4min) $opts(q4max) $opts(q4prob)"


       puts $f "\t\t \$qEC addPHBEntry 10 0 0"
       puts $f "\t\t \$qEC addPHBEntry 11 0 1"
       puts $f "\t\t \$qEC addPHBEntry 12 0 2"
       puts $f "\t\t \$qEC addPHBEntry 0 1 0"



      
       puts $f "\t \}"



       puts $f "\t foreach abc \{"


        set edgelist [$self GetNodes "diffcores"]

        foreach diedge [split $edgelist :] {
          set corelist [$self GetNodes "coretocore$diedge"]
          foreach dicore [split $corelist :] {
          puts $f "\t\t \{$diedge $dicore\}"
          }
        }

       puts $f "\t \} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"
       puts $f "\t\t \$qEC set numQueues_ 2"
       puts $f "\t\t \$qEC setNumPrec 3"


       puts $f "\t\t \$qEC configQ 0 0 $opts(q1min) $opts(q1max) $opts(q1prob)"
       puts $f "\t\t \$qEC configQ 0 1 $opts(q2min) $opts(q2max) $opts(q2prob)"
       puts $f "\t\t \$qEC configQ 0 2 $opts(q3min) $opts(q3max) $opts(q3prob)"
       puts $f "\t\t \$qEC configQ 1 0 $opts(q4min) $opts(q4max) $opts(q4prob)"


       puts $f "\t\t \$qEC addPHBEntry 10 0 0"
       puts $f "\t\t \$qEC addPHBEntry 11 0 1"
       puts $f "\t\t \$qEC addPHBEntry 12 0 2"
       puts $f "\t\t \$qEC addPHBEntry 0 1 0"
 
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 1\]\) \$no\(\[lindex \$abc 0\]\)\] queue\]"
       puts $f "\t\t \$qEC set numQueues_ 2"
       puts $f "\t\t \$qEC setNumPrec 3"
       
       puts $f "\t\t \$qEC configQ 0 0 $opts(q1min) $opts(q1max) $opts(q1prob)"
       puts $f "\t\t \$qEC configQ 0 1 $opts(q2min) $opts(q2max) $opts(q2prob)"
       puts $f "\t\t \$qEC configQ 0 2 $opts(q3min) $opts(q3max) $opts(q3prob)"
       puts $f "\t\t \$qEC configQ 1 0 $opts(q4min) $opts(q4max) $opts(q4prob)"

       puts $f "\t\t \$qEC addPHBEntry 10 0 0"
       puts $f "\t\t \$qEC addPHBEntry 11 0 1"
       puts $f "\t\t \$qEC addPHBEntry 12 0 2"   
       puts $f "\t\t \$qEC addPHBEntry 0 1 0"
       

       puts $f "\t \}"

    


       puts $f "\}"


# print out the Statictics at all links

       puts $f "proc out-pol \{ sim node \} \{ "
       puts $f "puts \"in policy...........\""
       puts $f "\t upvar \$sim ns"
       puts $f "\t upvar \$node no"
       puts $f "\t foreach abc \{"


       set edgelist [$self GetNodes "diffcores"]

       foreach diedge [split $edgelist :] {
         set corelist [$self GetNodes "coretocore$diedge"]
          foreach dicore [split $corelist :] {
          puts $f "\t\t \{$diedge $dicore\}"
          }
        }

       puts $f "\t \} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"
       puts $f "\t\t \$qEC printStats"
       
       puts $f "\t \}"


       puts $f "\t foreach abc \{"
       
        set edgelist [$self GetNodes "diffedges"]

        foreach diedge [split $edgelist :] {
          set corelist [$self GetNodes "edgetocore$diedge"]
          foreach dicore [split $corelist :] {
          puts $f "\t\t \{$dicore $diedge\}"
          }
        }

       puts $f "\t \} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"
       puts $f "\t\t \$qEC printStats"
       puts $f "\t \}"



       puts $f "\t foreach abc \{"

        set edgelist [$self GetNodes "diffedges"]

        foreach diedge [split $edgelist :] {
          set corelist [$self GetNodes "edgetocore$diedge"]
          foreach dicore [split $corelist :] {
          puts $f "\t\t \{$diedge $dicore\}"
          }
        }

       puts $f "\t \} \{"
       puts $f "\t\t set qEC \[\[\$ns link \$no\(\[lindex \$abc 0\]\) \$no\(\[lindex \$abc 1\]\)\] queue\]"
       puts $f "\t\t \$qEC printStats"

       puts $f "\t \}"



       puts $f "\}"
 
       
       flush $f
       close $f
}




